package com.minemaarten.signals.block;

import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import net.minecraft.block.BlockRailBase;
import net.minecraft.block.BlockRailPowered;
import net.minecraft.block.ITileEntityProvider;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.properties.PropertyBool;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityList;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.item.EntityMinecartContainer;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.server.MinecraftServer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumHand;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.text.TextComponentTranslation;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;
import net.minecraftforge.common.ForgeHooks;
import net.minecraftforge.fml.common.registry.GameRegistry;

import com.minemaarten.signals.client.CreativeTabSignals;
import com.minemaarten.signals.init.ModBlocks;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketSpawnParticle;
import com.minemaarten.signals.rail.network.NetworkRail;
import com.minemaarten.signals.rail.network.RailEdge;
import com.minemaarten.signals.rail.network.mc.MCPos;
import com.minemaarten.signals.rail.network.mc.MCTrain;
import com.minemaarten.signals.rail.network.mc.RailNetworkManager;
import com.minemaarten.signals.tileentity.TileEntityTeleportRail;

import net.minecraft.block.BlockRailBase.EnumRailDirection;

public class BlockTeleportRail extends BlockRailBase implements ITileEntityProvider{

    /**
     * Together with the SHAPE, this determines which way carts will go after they get teleported to this teleport rail.
     */
    public static final PropertyBool FORWARD = PropertyBool.func_177716_a("forward");

    public BlockTeleportRail(){
        super(false);
        func_149663_c("teleport_rail");
        func_149647_a(CreativeTabSignals.getInstance());
        ModBlocks.registerBlock(this);
        GameRegistry.registerTileEntity(TileEntityTeleportRail.class, "signals:teleport_rail");
    }

    @Override
    public boolean isFlexibleRail(IBlockAccess world, BlockPos pos){
        return false;
    }

    @Override
    public void onMinecartPass(World world, EntityMinecart cart, BlockPos pos){
        super.onMinecartPass(world, cart, pos);
        if(!world.field_72995_K) {

            TileEntityTeleportRail te = (TileEntityTeleportRail)world.func_175625_s(pos);
            MCPos telDestination = te.getLinkedPosition();
            if(telDestination != null) { //We can only teleport if we have a destination to go to.
                spawnParticle(cart);
                MCTrain train = RailNetworkManager.getServerInstance().getState().getTrain(cart.func_110124_au());
                if(train != null) { //Should always be true
                    if(train.getCarts().size() == train.cartIDs.size()) { //We can only teleport if all carts are loaded.
                        //Check if all carts are on adjacent teleport tracks
                        Set<MCPos> teleportRails = getNeighboringTeleportRails(world, pos);
                        for(MCPos cartPos : train.getPositions()) {
                            if(!teleportRails.contains(cartPos)) {
                                return; //When not all carts are on adjacent teleport tracks, don't teleport.
                            }
                        }

                        teleport(new HashSet<>(train.getCarts()), telDestination);
                    }
                }
            }
        }
    }

    private void teleport(Set<EntityMinecart> carts, MCPos destination){
        for(EntityMinecart cart : carts) {
            teleport(cart, destination);
        }
    }

    private void teleport(EntityMinecart cart, MCPos destination){
        if(cart instanceof EntityMinecartContainer) {
            ((EntityMinecartContainer)cart).field_94112_b = false;
        }

        int dimensionIn = destination.getDimID();
        BlockPos destPos = destination.getPos();
        if(!ForgeHooks.onTravelToDimension(cart, destination.getDimID())) return;

        MinecraftServer minecraftserver = cart.func_184102_h();
        int i = cart.field_71093_bK;
        WorldServer worldserver = minecraftserver.func_71218_a(i);
        WorldServer worldserver1 = minecraftserver.func_71218_a(dimensionIn);
        cart.field_71093_bK = dimensionIn;

        /*if (i == 1 && dimensionIn == 1)
        {
            worldserver1 = minecraftserver.getWorld(0);
            this.dimension = 0;
        }*/

        cart.field_70170_p.func_72900_e(cart);
        cart.field_70128_L = false;
        BlockPos blockpos = destination.getPos();

        /* double moveFactor = worldserver.provider.getMovementFactor() / worldserver1.provider.getMovementFactor();
        double d0 = MathHelper.clamp(this.posX * moveFactor, worldserver1.getWorldBorder().minX() + 16.0D, worldserver1.getWorldBorder().maxX() - 16.0D);
        double d1 = MathHelper.clamp(this.posZ * moveFactor, worldserver1.getWorldBorder().minZ() + 16.0D, worldserver1.getWorldBorder().maxZ() - 16.0D);
        double d2 = 8.0D;*/

        cart.func_174828_a(destPos, 90.0F, 0.0F);
        /*Teleporter teleporter = worldserver1.getDefaultTeleporter();
        teleporter.placeInExistingPortal(this, f);
        blockpos = new BlockPos(this);*/

        worldserver.func_72866_a(cart, false);
        Entity entity = EntityList.func_191304_a(cart.getClass(), worldserver1);

        if(entity != null) {
            copyDataFromOld(entity, cart);

            entity.func_174828_a(blockpos, entity.field_70177_z, entity.field_70125_A);

            IBlockState state = destination.getLoadedBlockState();
            if(state.func_177230_c() == ModBlocks.TELEPORT_RAIL) { //If the destination is a teleport track, use its rail direction to push the cart the right direction
                EnumFacing teleportDir = getTeleportDirection(state);
                double speed = 0.2;
                entity.field_70159_w = teleportDir.func_82601_c() * speed;
                entity.field_70181_x = teleportDir.func_96559_d() * speed;
                entity.field_70179_y = teleportDir.func_82599_e() * speed;
            } else {
                entity.field_70159_w = entity.field_70181_x = entity.field_70179_y = 0;
            }

            boolean flag = entity.field_98038_p;
            entity.field_98038_p = true;
            worldserver1.func_72838_d(entity);
            entity.field_98038_p = flag;
            worldserver1.func_72866_a(entity, false);
        }

        cart.field_70128_L = true;
        worldserver.func_82742_i();
        worldserver1.func_82742_i();
    }

    /**
     * Prepares this entity in new dimension by copying NBT data from entity in old dimension
     */
    private static void copyDataFromOld(Entity newEntity, Entity entityIn){
        NBTTagCompound nbttagcompound = entityIn.func_189511_e(new NBTTagCompound());
        nbttagcompound.func_82580_o("Dimension");
        newEntity.func_70020_e(nbttagcompound);
    }

    private Set<MCPos> getNeighboringTeleportRails(World world, BlockPos pos){
        Set<MCPos> ret = new HashSet<MCPos>();
        MCPos mcPos = new MCPos(world, pos);
        RailEdge<MCPos> edge = RailNetworkManager.getServerInstance().getNetwork().findEdge(mcPos);
        int startIndex = edge.getIndex(mcPos);
        Object thisRailType = edge.get(startIndex).getRailType();

        // travel backwards
        int curIndex = startIndex;
        while(curIndex >= 0) {
            NetworkRail<MCPos> neighborRail = edge.get(curIndex);
            if(!neighborRail.getRailType().equals(thisRailType) || neighborRail.getPos().getDimID() != world.field_73011_w.getDimension()) break;
            ret.add(neighborRail.getPos());
            curIndex--;
        }

        //travel forwards
        curIndex = startIndex + 1;
        while(curIndex < edge.length) {
            NetworkRail<MCPos> neighborRail = edge.get(curIndex);
            if(!neighborRail.getRailType().equals(thisRailType) || neighborRail.getPos().getDimID() != world.field_73011_w.getDimension()) break;
            ret.add(neighborRail.getPos());
            curIndex++;
        }

        return ret;
    }

    private void spawnParticle(EntityMinecart cart){
        Random rand = cart.field_70170_p.field_73012_v;
        float maxSpeed = 1;
        float f = (rand.nextFloat() - 0.5F) * maxSpeed;
        float f1 = (rand.nextFloat() - 0.5F) * maxSpeed;
        float f2 = (rand.nextFloat() - 0.5F) * maxSpeed;
        NetworkHandler.sendToAllAround(new PacketSpawnParticle(EnumParticleTypes.PORTAL, cart.field_70165_t, cart.field_70163_u, cart.field_70161_v, f, f1, f2), cart.field_70170_p);
    }

    @Override
    public IProperty<EnumRailDirection> func_176560_l(){
        return BlockRailPowered.field_176568_b;
    }

    @Override
    protected BlockStateContainer func_180661_e(){
        return new BlockStateContainer(this, func_176560_l(), FORWARD);
    }

    /**
     * Convert the given metadata into a BlockState for this Block
     */
    @Override
    public IBlockState func_176203_a(int meta){
        return this.func_176223_P().func_177226_a(func_176560_l(), BlockRailBase.EnumRailDirection.func_177016_a(meta >> 1)).func_177226_a(FORWARD, (meta & 1) > 0);
    }

    /**
     * Convert the BlockState into the correct metadata value
     */
    @Override
    public int func_176201_c(IBlockState state){
        return (state.func_177229_b(func_176560_l()).func_177015_a() << 1) + (state.func_177229_b(FORWARD) ? 1 : 0);
    }

    @Override
    public TileEntity func_149915_a(World worldIn, int meta){
        return new TileEntityTeleportRail();
    }

    /**
     * Called serverside after this block is replaced with another in Chunk, but before the Tile Entity is updated
     */
    @Override
    public void func_180663_b(World worldIn, BlockPos pos, IBlockState state){
        super.func_180663_b(worldIn, pos, state);
        worldIn.func_175713_t(pos);
    }

    private EnumFacing getTeleportDirection(IBlockState state){
        EnumRailDirection railDir = state.func_177229_b(func_176560_l());
        boolean forward = state.func_177229_b(FORWARD);
        return getDirection(railDir, forward);
    }

    public static EnumFacing getDirection(EnumRailDirection railDir, boolean forward){
        switch(railDir){
            case EAST_WEST:
                return forward ? EnumFacing.EAST : EnumFacing.WEST;
            case ASCENDING_EAST:
                return forward ? EnumFacing.EAST : EnumFacing.WEST;
            case ASCENDING_WEST:
                return forward ? EnumFacing.WEST : EnumFacing.EAST;
            case NORTH_SOUTH:
                return forward ? EnumFacing.NORTH : EnumFacing.SOUTH;
            case ASCENDING_NORTH:
                return forward ? EnumFacing.NORTH : EnumFacing.SOUTH;
            case ASCENDING_SOUTH:
                return forward ? EnumFacing.SOUTH : EnumFacing.NORTH;
            default:
                throw new IllegalStateException("Invalid rail dir for teleport track: " + railDir);
        }
    }

    @Override
    public boolean func_180639_a(World worldIn, BlockPos pos, IBlockState state, EntityPlayer playerIn, EnumHand hand, EnumFacing facing, float hitX, float hitY, float hitZ){
        if(playerIn.func_70093_af()) {
            if(!worldIn.field_72995_K) {
                worldIn.func_175656_a(pos, state.func_177231_a(FORWARD));
                playerIn.func_145747_a(new TextComponentTranslation("signals.message.teleport_rail_toggled_direction"));
            }
            return true;
        } else {
            return false;
        }
    }
}
