package com.minemaarten.signals.block;

import net.minecraft.block.BlockRailPowered;
import net.minecraft.block.ITileEntityProvider;
import net.minecraft.block.properties.PropertyBool;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumHand;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.util.text.TextComponentTranslation;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.fml.common.registry.GameRegistry;

import com.minemaarten.signals.client.CreativeTabSignals;
import com.minemaarten.signals.init.ModBlocks;
import com.minemaarten.signals.tileentity.TileEntityTransportRail;

public class BlockTransportRail extends BlockRailPowered implements ITileEntityProvider{
    /**
     * Together with the SHAPE, this determines which way carts will go when standing still and then getting powered.
     */
    public static final PropertyBool FORWARD = PropertyBool.func_177716_a("forward");

    public BlockTransportRail(){
        func_149663_c("transport_rail");
        func_149647_a(CreativeTabSignals.getInstance());
        ModBlocks.registerBlock(this);
        GameRegistry.registerTileEntity(TileEntityTransportRail.class, "transport_rail");
    }

    @Override
    public void onMinecartPass(World world, EntityMinecart cart, BlockPos pos){
        super.onMinecartPass(world, cart, pos);
        Vec3d vec = new Vec3d(cart.field_70159_w, cart.field_70181_x, cart.field_70179_y);
        double maxAcceleration = 0.2D;
        double vecLength = vec.func_72433_c();
        IBlockState state = world.func_180495_p(pos);
        state = state.func_177230_c().func_176221_a(state, world, pos);
        boolean powered = state.func_177229_b(field_176569_M);
        double maxSpeed = powered ? 0.1D : 0.0D;
        if(vecLength > maxSpeed) {
            vecLength = Math.max(vecLength - maxAcceleration, maxSpeed);
        } else if(vecLength < maxSpeed) {
            if(vecLength < 0.1) { //When practically standing still
                EnumFacing pushDir = BlockTeleportRail.getDirection(state.func_177229_b(func_176560_l()), state.func_177229_b(FORWARD));
                vec = new Vec3d(pushDir.func_82601_c(), 0, pushDir.func_82599_e()); //Push in the direction of the rail.
            }
            vecLength = Math.min(vecLength + maxAcceleration, maxSpeed);
        }
        vec = vec.func_72432_b().func_186678_a(vecLength);
        cart.field_70159_w = vec.field_72450_a;
        cart.field_70181_x = vec.field_72448_b;
        cart.field_70179_y = vec.field_72449_c;
    }

    @Override
    public TileEntity func_149915_a(World worldIn, int meta){
        return new TileEntityTransportRail();
    }

    /**
     * Called serverside after this block is replaced with another in Chunk, but before the Tile Entity is updated
     */
    @Override
    public void func_180663_b(World worldIn, BlockPos pos, IBlockState state){
        super.func_180663_b(worldIn, pos, state);
        worldIn.func_175713_t(pos);
    }

    @Override
    protected BlockStateContainer func_180661_e(){
        return new BlockStateContainer(this, func_176560_l(), field_176569_M, FORWARD);
    }

    @Override
    public IBlockState func_176221_a(IBlockState state, IBlockAccess worldIn, BlockPos pos){
        TileEntityTransportRail te = (TileEntityTransportRail)worldIn.func_175625_s(pos);
        return super.func_176221_a(state, worldIn, pos).func_177226_a(FORWARD, te.isForward());
    }

    @Override
    public boolean func_180639_a(World worldIn, BlockPos pos, IBlockState state, EntityPlayer playerIn, EnumHand hand, EnumFacing facing, float hitX, float hitY, float hitZ){
        if(playerIn.func_70093_af()) {
            if(!worldIn.field_72995_K) {
                TileEntityTransportRail te = (TileEntityTransportRail)worldIn.func_175625_s(pos);
                te.toggleForward();
                playerIn.func_145747_a(new TextComponentTranslation("signals.message.teleport_rail_toggled_direction"));
            }
            return true;
        } else {
            return false;
        }
    }
}
