package com.minemaarten.signals.client.gui;

import java.awt.Point;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.FontRenderer;
import net.minecraft.client.gui.GuiButton;
import net.minecraft.client.gui.ScaledResolution;
import net.minecraft.client.gui.inventory.GuiContainer;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.resources.I18n;
import net.minecraft.inventory.Container;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.Slot;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.ResourceLocation;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import org.apache.commons.lang3.text.WordUtils;
import org.lwjgl.opengl.GL11;

import com.minemaarten.signals.Signals;
import com.minemaarten.signals.client.gui.widget.IGuiWidget;
import com.minemaarten.signals.client.gui.widget.IWidgetListener;
import com.minemaarten.signals.client.gui.widget.WidgetLabel;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketGuiButton;

@SideOnly(Side.CLIENT)
public class GuiContainerBase<Tile extends TileEntity> extends GuiContainer implements IWidgetListener{

    public final Tile te;
    private final ResourceLocation guiTexture;
    protected final List<IGuiWidget> widgets = new ArrayList<>();
    private boolean refreshScheduled;
    private boolean hasInit; //Fix for some weird race condition occuring in 1.8 where drawing is called before initGui().

    public GuiContainerBase(Container par1Container, Tile te, String guiTexture){
        super(par1Container);
        this.te = te;
        this.guiTexture = guiTexture != null ? new ResourceLocation(guiTexture) : null;
    }

    protected void addWidget(IGuiWidget widget){
        widgets.add(widget);
        widget.setListener(this);
    }

    protected void addWidgets(Iterable<IGuiWidget> widgets){
        for(IGuiWidget widget : widgets) {
            addWidget(widget);
        }
    }

    protected void addLabel(String text, int x, int y){
        addWidget(new WidgetLabel(x, y, text));
    }

    protected void removeWidget(IGuiWidget widget){
        widgets.remove(widget);
    }

    @Override
    public void func_73866_w_(){
        super.func_73866_w_();
        hasInit = true;
    }

    @Override
    protected void func_146976_a(float partialTicks, int i, int j){
        if(shouldDrawBackground()) {
            GL11.glColor4f(1.0F, 1.0F, 1.0F, 1.0F);
            bindGuiTexture();
            int xStart = (field_146294_l - field_146999_f) / 2;
            int yStart = (field_146295_m - field_147000_g) / 2;
            func_73729_b(xStart, yStart, 0, 0, field_146999_f, field_147000_g);
        }

        GL11.glColor4d(1, 1, 1, 1);
        GL11.glDisable(GL11.GL_LIGHTING);
        for(IGuiWidget widget : widgets) {
            widget.render(i, j, partialTicks);
        }
        for(IGuiWidget widget : widgets) {
            widget.postRender(i, j, partialTicks);
        }
    }

    /**
     * Draws dark gray back layer with inventory spots.
     */
    protected void drawBackLayer(){
        GlStateManager.func_179090_x();
        int borderWidth = 1;
        func_73734_a(field_147003_i - borderWidth, field_147009_r - borderWidth, field_147003_i + field_146999_f + borderWidth, field_147009_r + field_147000_g + borderWidth, 0xFF222222);
        func_73734_a(field_147003_i, field_147009_r, field_147003_i + field_146999_f, field_147009_r + field_147000_g, 0xFF333333);
        for(Slot slot : field_147002_h.field_75151_b) {
            func_73734_a(field_147003_i + slot.field_75223_e, field_147009_r + slot.field_75221_f, field_147003_i + slot.field_75223_e + 16, field_147009_r + slot.field_75221_f + 16, 0xFF222222);
        }
        GlStateManager.func_179098_w();
    }

    protected void drawDarkGreyTextCentered(int x, int y, String localizationKey, Object... args){
        x -= field_146289_q.func_78256_a(I18n.func_135052_a(localizationKey, args)) / 2 + 1;
        drawDarkGreyText(x, y, localizationKey, args);
    }

    protected void drawDarkGreyText(int x, int y, String localizationKey, Object... args){
        String text = I18n.func_135052_a(localizationKey, args);
        func_73734_a(x, y, x + field_146289_q.func_78256_a(text) + 3, y + field_146289_q.field_78288_b + 3, 0xFF222222);
        field_146289_q.func_78276_b(text, x + 2, y + 2, 0xFFFFFFFF);
    }

    protected boolean shouldDrawBackground(){
        return true;
    }

    protected void bindGuiTexture(){
        if(guiTexture != null) {
            field_146297_k.func_110434_K().func_110577_a(guiTexture);
            GL11.glEnable(GL11.GL_TEXTURE_2D);
        }
    }

    @Override
    protected void func_146979_b(int x, int y){
        if(getInvNameOffset() != null && te instanceof IInventory) {
            IInventory inv = (IInventory)te;
            String containerName = inv.func_145818_k_() ? inv.func_70005_c_() : I18n.func_135052_a(inv.func_70005_c_() + ".name");
            field_146289_q.func_78276_b(containerName, field_146999_f / 2 - field_146289_q.func_78256_a(containerName) / 2 + getInvNameOffset().x, 6 + getInvNameOffset().y, 4210752);
        }
        if(getInvTextOffset() != null) field_146289_q.func_78276_b(I18n.func_135052_a("container.inventory"), 8 + getInvTextOffset().x, field_147000_g - 94 + getInvTextOffset().y, 4210752);
    }

    protected Point getInvNameOffset(){
        return new Point(0, 0);
    }

    protected Point getInvTextOffset(){
        return new Point(0, 0);
    }

    @Override
    public void func_73863_a(int x, int y, float partialTick){
        if(!hasInit) return;
        super.func_73863_a(x, y, partialTick);

        List<String> tooltip = new ArrayList<>();

        GL11.glColor4d(1, 1, 1, 1);
        GL11.glDisable(GL11.GL_LIGHTING);
        for(IGuiWidget widget : widgets) {
            if(widget.getBounds().contains(x, y)) widget.addTooltip(x, y, tooltip, Signals.proxy.isSneakingInGui());
        }

        if(!tooltip.isEmpty()) {
            List<String> localizedTooltip = new ArrayList<>();
            for(String line : tooltip) {
                String localizedLine = I18n.func_135052_a(line);
                String[] lines = WordUtils.wrap(localizedLine, 50).split(System.getProperty("line.separator"));
                for(String locLine : lines) {
                    localizedTooltip.add(locLine);
                }
            }
            drawHoveringText(localizedTooltip, x, y, field_146289_q);
        }
    }

    @Override
    public void func_73876_c(){
        if(refreshScheduled) {
            refreshScheduled = false;
            refreshScreen();
        }

        super.func_73876_c();

        for(IGuiWidget widget : widgets)
            widget.update();
    }

    @Override
    protected void func_146284_a(GuiButton button){
        sendPacketToServer(button.field_146127_k);
    }

    @Override
    protected void func_73864_a(int par1, int par2, int par3) throws IOException{
        super.func_73864_a(par1, par2, par3);
        for(IGuiWidget widget : widgets) {
            if(widget.getBounds().contains(par1, par2)) widget.onMouseClicked(par1, par2, par3);
            else widget.onMouseClickedOutsideBounds(par1, par2, par3);
        }
    }

    @Override
    public void actionPerformed(IGuiWidget widget){
        sendPacketToServer(widget.getID());
    }

    protected void sendPacketToServer(int id){
        NetworkHandler.sendToServer(new PacketGuiButton(id));
    }

    @Override
    public void func_146274_d() throws IOException{
        super.func_146274_d();
        for(IGuiWidget widget : widgets) {
            widget.handleMouseInput();
        }
    }

    @Override
    protected void func_73869_a(char key, int keyCode) throws IOException{
        for(IGuiWidget widget : widgets) {
            if(widget.onKey(key, keyCode)) return;
        }
        super.func_73869_a(key, keyCode);
    }

    @Override
    public void func_146280_a(Minecraft par1Minecraft, int par2, int par3){
        widgets.clear();
        super.func_146280_a(par1Minecraft, par2, par3);
    }

    public void refreshScreen(){
        ScaledResolution scaledresolution = new ScaledResolution(field_146297_k);
        int i = scaledresolution.func_78326_a();
        int j = scaledresolution.func_78328_b();
        func_146280_a(field_146297_k, i, j);
    }

    public void scheduleRefresh(){
        refreshScheduled = true;
    }

    public void drawHoveringString(List<String> text, int x, int y, FontRenderer fontRenderer){
        drawHoveringText(text, x, y, fontRenderer);
    }

    /* public static void drawTexture(String texture, int x, int y){
         Minecraft.getMinecraft().getTextureManager().bindTexture(GuiUtils.getResourceLocation(texture));
         VertexBuffer wr = Tessellator.getInstance().getBuffer();
         wr.begin(GL11.GL_QUADS, DefaultVertexFormats.POSITION_TEX);
         wr.pos(x, y + 16, 0).tex(0.0, 1.0).endVertex();
         wr.pos(x + 16, y + 16, 0).tex(1.0, 1.0).endVertex();
         wr.pos(x + 16, y, 0).tex(1.0, 0.0).endVertex();
         wr.pos(x, y, 0).tex(0.0, 0.0).endVertex();
         Tessellator.getInstance().draw();
         // this.drawTexturedModalRect(x, y, 0, 0, 16, 16);
     }*/

    @Override
    public int getGuiLeft(){
        return field_147003_i;
    }

    @Override
    public int getGuiTop(){
        return field_147009_r;
    }

    @Override
    public void onKeyTyped(IGuiWidget widget){

    }

    public void onFieldSynced(){

    }
}
