package com.minemaarten.signals.client.gui;

import java.awt.Point;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import net.minecraft.inventory.Container;
import net.minecraft.tileentity.TileEntity;

import com.minemaarten.signals.api.access.IDestinationAccessor;
import com.minemaarten.signals.client.gui.widget.IGuiWidget;
import com.minemaarten.signals.client.gui.widget.WidgetComboBox;
import com.minemaarten.signals.rail.network.mc.RailNetworkManager;

public abstract class GuiDestinations<DestinationAccessor extends IDestinationAccessor> extends
        GuiContainerBase<TileEntity>{

    private WidgetComboBox[] stationNameFields;
    protected final DestinationAccessor destinationAccessor;

    public GuiDestinations(Container container, DestinationAccessor destinationAccessor){
        super(container, null, null);
        this.destinationAccessor = destinationAccessor;
        field_147000_g = 200;
        field_146999_f = 120;
    }

    @Override
    public void func_73866_w_(){
        super.func_73866_w_();
        WidgetComboBox[] oldFields = stationNameFields;
        stationNameFields = new WidgetComboBox[destinationAccessor.getTotalDestinations() + 1];
        for(int i = 0; i < stationNameFields.length; i++) {
            stationNameFields[i] = new WidgetComboBox(field_146289_q, field_147003_i + 10, field_146295_m / 2 - 5 - stationNameFields.length * 7 + i * 14, 100, field_146289_q.field_78288_b);
            stationNameFields[i].setElements(RailNetworkManager.getClientInstance().getNetwork().getStationNames());
            addWidget(stationNameFields[i]);
        }
        if(oldFields != null) {
            for(int i = 0; i < oldFields.length && i < stationNameFields.length; i++) {
                if(oldFields[i].func_146206_l()) stationNameFields[i].func_146180_a(oldFields[i].func_146179_b());
                stationNameFields[i].func_146195_b(oldFields[i].func_146206_l());
            }
        }
    }

    @Override
    protected boolean shouldDrawBackground(){
        return false;
    }

    @Override
    protected Point getInvTextOffset(){
        return null;
    }

    protected boolean showDestination(){
        return true;
    }

    @Override
    protected void func_146976_a(float partialTicks, int i, int j){
        drawBackLayer();

        if(showDestination()) {
            String dest = destinationAccessor.getCurrentDestination();
            if(dest.equals("")) {
                drawDarkGreyTextCentered(field_147003_i + 60, field_147009_r + field_147000_g - 18, "signals.gui.cart.no_destination");
            } else {
                drawDarkGreyTextCentered(field_147003_i + 60, field_147009_r + field_147000_g - 18, "signals.gui.cart.destination", dest);
            }
        }

        if(destinationAccessor.getTotalDestinations() > 0) {
            if(showDestination()) drawDestinationRect(destinationAccessor.getDestinationIndex(), 0xFF005500);
            for(int invalidDest : destinationAccessor.getInvalidDestinationIndeces()) {
                drawDestinationRect(invalidDest, 0xAAFF0000);
            }
        }

        super.func_146976_a(partialTicks, i, j);
    }

    private void drawDestinationRect(int destinationIndex, int color){
        int indicatorX = field_147003_i + 7;
        int indicatorY = field_146295_m / 2 - stationNameFields.length * 7 + destinationIndex * 14 - 8;
        func_73734_a(indicatorX, indicatorY, indicatorX + 106, indicatorY + field_146289_q.field_78288_b + 6, color);
    }

    @Override
    public void func_73876_c(){
        super.func_73876_c();
        String[] destinations = destinationAccessor.getDestinations();
        for(int i = 0; i < stationNameFields.length; i++) {
            if(!stationNameFields[i].func_146206_l()) {
                stationNameFields[i].func_146180_a(i < destinations.length ? destinations[i] : "");
            }
        }
        if(destinationAccessor.getTotalDestinations() != stationNameFields.length - 1) scheduleRefresh();
    }

    @Override
    public void onKeyTyped(IGuiWidget widget){
        super.onKeyTyped(widget);

        destinationAccessor.setDestinations(getDestinations());
    }

    private List<String> getDestinations(){
        return Stream.of(stationNameFields).map(widget -> widget.func_146179_b()).collect(Collectors.toList());
    }
}
