package com.minemaarten.signals.rail;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.Gui;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.texture.DynamicTexture;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import org.lwjgl.opengl.GL11;

import com.google.common.base.Predicates;
import com.minemaarten.signals.capabilities.CapabilityMinecartDestination;
import com.minemaarten.signals.rail.network.INetworkObject;
import com.minemaarten.signals.rail.network.NetworkStation;
import com.minemaarten.signals.rail.network.RailNetwork;
import com.minemaarten.signals.rail.network.RailNetworkClient;
import com.minemaarten.signals.rail.network.RailRoute;
import com.minemaarten.signals.rail.network.Train;
import com.minemaarten.signals.rail.network.mc.MCPos;
import com.minemaarten.signals.rail.network.mc.RailNetworkManager;

public class NetworkController{
    public static final int RAIL_COLOR = 0xFF666666;
    public static final int PATH_COLOR = 0xFFAAAAAA;
    public static final int TEXT_COLOR = 0xFFFFFF; //No alpha
    public static final int STATION_COLOR = 0xFFDDDD00;
    public static final int NOTHING_COLOR = 0xFF222222;
    private static Map<Integer, NetworkController> cache = new HashMap<>();
    private static RailNetworkClient<MCPos> network = RailNetworkClient.empty();

    public static NetworkController getInstance(World world){
        if(!world.field_72995_K) throw new IllegalStateException("Can only be called client side!");
        return getInstance(world.field_73011_w.getDimension());
    }

    public static NetworkController getInstance(int dimension){
        if(network != RailNetworkManager.getClientInstance().getClientNetwork()) {
            network = RailNetworkManager.getClientInstance().getClientNetwork();
            cache = rebuildAll();
        }

        NetworkController controller = cache.get(dimension);
        if(controller == null) {
            controller = new NetworkController(dimension);
            controller.setColors(new int[]{0}, 1, 1, 0, 0);
            cache.put(dimension, controller);
        }
        return controller;
    }

    @SideOnly(Side.CLIENT)
    private DynamicTexture dynamicTexture;
    private int width, height;
    private int[] colors;
    private int startX, startZ;
    private final int dimensionId;
    private ResourceLocation textureLoc;

    public NetworkController(int dimensionId){
        this.dimensionId = dimensionId;
    }

    @SideOnly(Side.CLIENT)
    public void setColors(int[] colors, int width, int height, int startX, int startZ){
        this.startX = startX;
        this.startZ = startZ;
        if(dynamicTexture == null || this.width != width || this.height != height) {
            this.width = width;
            this.height = height;
            if(dynamicTexture != null) dynamicTexture.func_147631_c();
            dynamicTexture = new DynamicTexture(width, height);
            textureLoc = Minecraft.func_71410_x().func_110434_K().func_110578_a("networkController/" + dimensionId, dynamicTexture);
        }
        System.arraycopy(colors, 0, dynamicTexture.func_110565_c(), 0, colors.length);
        dynamicTexture.func_110564_a();
    }

    public int getWidth(){
        return width;
    }

    public int getHeight(){
        return height;
    }

    public int getStartX(){
        return startX;
    }

    public int getStartZ(){
        return startZ;
    }

    private List<RailRoute<MCPos>> getAllRoutes(){
        return RailNetworkManager.getClientInstance().getAllTrains().map(t -> t.getCurRoute()).filter(Predicates.notNull()).collect(Collectors.toList());
    }

    @SideOnly(Side.CLIENT)
    public void render(World world){
        if(textureLoc == null) setColors(new int[1], 1, 1, 0, 0);
        Minecraft mc = Minecraft.func_71410_x();
        mc.func_110434_K().func_110577_a(textureLoc);
        GL11.glEnable(GL11.GL_TEXTURE_2D);
        GL11.glColor4f(1.0F, 1.0F, 1.0F, 1.0F);
        Gui.func_146110_a(0, 0, 0, 0, width, height, width, height);
        Tessellator t = Tessellator.func_178181_a();
        BufferBuilder buffer = t.func_178180_c();

        //Draw the path
        for(RailRoute<MCPos> route : getAllRoutes()) {
            for(MCPos pathPos : route.routeRails) {
                if(pathPos.getDimID() == dimensionId) {
                    int pathX = pathPos.getX() - startX;
                    int pathZ = pathPos.getZ() - startZ;
                    Gui.func_73734_a(pathX, pathZ, pathX + 1, pathZ + 1, PATH_COLOR);
                }
            }
        }

        //Draw the train locations
        GlStateManager.func_179124_c(0, 0, 1);
        GlStateManager.func_179090_x();
        buffer.func_181668_a(GL11.GL_QUADS, DefaultVertexFormats.field_181705_e);
        for(Train<MCPos> train : RailNetworkManager.getClientInstance().getAllTrains().collect(Collectors.toList())) {
            for(MCPos pos : train.getPositions()) {
                if(pos.getDimID() != dimensionId) continue;

                double x = pos.getX() - startX;
                double y = pos.getZ() - startZ;

                buffer.func_181662_b(x, y, 0).func_181675_d();
                buffer.func_181662_b(x, y + 1, 0).func_181675_d();
                buffer.func_181662_b(x + 1, y + 1, 0).func_181675_d();
                buffer.func_181662_b(x + 1, y, 0).func_181675_d();
            }
        }
        t.func_78381_a();
        GlStateManager.func_179098_w();

        //Draw the destination names next to the trains
        for(EntityMinecart cart : world.func_72872_a(EntityMinecart.class, new AxisAlignedBB(startX, 0, startZ, startX + width, 255, startZ + height))) {
            double x = cart.field_70165_t - startX - 0.5;
            double y = cart.field_70161_v - startZ - 0.5;
            GlStateManager.func_179124_c(0, 0, 1);

            String dest = cart.getCapability(CapabilityMinecartDestination.INSTANCE, null).getCurrentDestination();
            if(dest != null) {
                GlStateManager.func_179094_E();
                double scale = 1 / 4D;
                GlStateManager.func_179137_b(x + 1.5, y, 0);
                GlStateManager.func_179139_a(scale, scale, scale);
                mc.field_71466_p.func_78276_b(dest, 0, 0, TEXT_COLOR);
                GlStateManager.func_179121_F();
            }
        }

        //Draw the station names next to the stations
        RailNetwork<MCPos> network = RailNetworkManager.getClientInstance().getNetwork();
        for(NetworkStation<MCPos> station : network.railObjects.getStations()) {
            double x = station.getPos().getX() - startX - 0.5;
            double y = station.getPos().getZ() - startZ - 0.5;

            GlStateManager.func_179094_E();
            double scale = 1 / 4D;
            GlStateManager.func_179137_b(x + 2, y, 0);
            GlStateManager.func_179139_a(scale, scale, scale);
            mc.field_71466_p.func_78276_b(station.stationName, 0, 0, 0xFFFFFF00);
            GlStateManager.func_179121_F();
        }
    }

    private void setColor(int x, int z, int color){
        setColor(x, z, color, true);
    }

    private void setColor(int x, int z, int color, boolean sendPacket){
        int index = (x - startX) + (z - startZ) * width;
        if(index < 0 || index >= colors.length) { //If we are out of bounds that means we don't properly know the right dims.
            rebuildAll();
        } else {
            colors[index] = color;
        }
    }

    public void updateColor(int color, BlockPos... positions){
        boolean rebuildAll = false;
        for(BlockPos pos : positions) {
            if(color == NOTHING_COLOR) {
                if(pos.func_177958_n() == startX || pos.func_177958_n() == startX + width || pos.func_177952_p() == startZ || pos.func_177952_p() == startZ + height) {
                    rebuildAll = true; //When a border rail is removed it might decrease the total map size.
                }
            } else {
                if(pos.func_177958_n() < startX || pos.func_177958_n() >= startX + width || pos.func_177952_p() < startZ || pos.func_177952_p() >= startZ + height) {
                    rebuildAll = true; //When the rail is outside the current map scope we need to increase the map size.
                }
            }
            if(!rebuildAll) {
                setColor(pos.func_177958_n(), pos.func_177952_p(), color);
                setColors(colors, width, height, startX, startZ);
            }
        }
        if(rebuildAll) rebuildAll();
    }

    public static Map<Integer, NetworkController> rebuildAll(){
        Map<Integer, NetworkController> cache = new HashMap<>();

        Collection<INetworkObject<MCPos>> allObjects = RailNetworkManager.getClientInstance().getNetwork().railObjects.getAllNetworkObjects().values();
        Map<Integer, List<INetworkObject<MCPos>>> objsByDim = allObjects.stream().collect(Collectors.groupingBy(o -> o.getPos().getDimID()));

        for(Map.Entry<Integer, List<INetworkObject<MCPos>>> entry : objsByDim.entrySet()) {
            NetworkController controller = new NetworkController(entry.getKey());
            controller.rebuild(entry.getValue());
            controller.setColors(controller.colors, controller.width, controller.height, controller.startX, controller.startZ);

            cache.put(entry.getKey(), controller);
        }

        return cache;
    }

    private void rebuild(Collection<INetworkObject<MCPos>> objects){
        startX = Integer.MAX_VALUE;
        startZ = Integer.MAX_VALUE;
        int endX = Integer.MIN_VALUE;
        int endZ = Integer.MIN_VALUE;
        for(INetworkObject<MCPos> obj : objects) {
            MCPos pos = obj.getPos();
            startX = Math.min(startX, pos.getX());
            startZ = Math.min(startZ, pos.getZ());
            endX = Math.max(endX, pos.getX());
            endZ = Math.max(endZ, pos.getZ());
        }

        width = endX - startX + 1;
        height = endZ - startZ + 1;
        colors = new int[width * height];
        Arrays.fill(colors, NOTHING_COLOR);

        for(INetworkObject<MCPos> obj : objects) {
            setColor(obj.getPos().getX(), obj.getPos().getZ(), obj.getColor(), false);
        }
    }
}
