package com.minemaarten.signals.tileentity;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import net.minecraft.block.state.IBlockState;
import net.minecraft.client.resources.I18n;
import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.play.server.SPacketUpdateTileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ITickable;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;

import com.minemaarten.signals.api.access.ISignal;
import com.minemaarten.signals.block.BlockSignalBase;
import com.minemaarten.signals.rail.network.INetworkObject;
import com.minemaarten.signals.rail.network.NetworkSignal;
import com.minemaarten.signals.rail.network.NetworkSignal.EnumSignalType;
import com.minemaarten.signals.rail.network.mc.MCPos;
import com.minemaarten.signals.rail.network.mc.RailNetworkManager;

import com.minemaarten.signals.api.access.ISignal.EnumForceMode;
import com.minemaarten.signals.api.access.ISignal.EnumLampStatus;

public abstract class TileEntitySignalBase extends TileEntityBase implements ITickable, ISignal{

    private boolean firstTick = true;
    private EnumForceMode forceMode = EnumForceMode.NONE;
    private String text = "";
    private String arguments = "";

    public EnumFacing getFacing(){
        IBlockState state = getBlockState();
        return state != null && state.func_177230_c() instanceof BlockSignalBase ? state.func_177229_b(BlockSignalBase.FACING) : EnumFacing.NORTH;
    }

    public abstract EnumSignalType getSignalType();

    public void setLampStatus(EnumLampStatus lampStatus){
        IBlockState state = getBlockState();
        if(state.func_177227_a().contains(BlockSignalBase.LAMP_STATUS) && state.func_177229_b(BlockSignalBase.LAMP_STATUS) != lampStatus) {
            func_145831_w().func_175656_a(func_174877_v(), state.func_177226_a(BlockSignalBase.LAMP_STATUS, lampStatus));
            if(lampStatus == EnumLampStatus.GREEN) {
                //Push carts when they're standing still or going backwards.
                List<EntityMinecart> neighborMinecarts = getNeighborMinecarts();
                for(EntityMinecart cart : neighborMinecarts) {
                    if(new Vec3d(cart.field_70159_w, cart.field_70181_x, cart.field_70179_y).func_72433_c() < 0.01 || EnumFacing.func_176737_a((float)cart.field_70159_w, 0, (float)cart.field_70179_y) == getFacing()) {
                        cart.field_70159_w += getFacing().func_82601_c() * 0.1;
                        cart.field_70179_y += getFacing().func_82599_e() * 0.1;
                    } else if(EnumFacing.func_176737_a((float)cart.field_70159_w, 0, (float)cart.field_70179_y) == getFacing().func_176734_d()) {
                        //Reverse the cart if going backwards into a signal.
                        cart.field_70159_w *= -1;
                        cart.field_70181_x *= -1;
                        cart.field_70179_y *= -1;
                    }
                }
            }
        }
    }

    @Override
    public EnumLampStatus getLampStatus(){
        if(func_145831_w() != null) {
            IBlockState state = func_145831_w().func_180495_p(func_174877_v());
            if(state.func_177227_a().contains(BlockSignalBase.LAMP_STATUS)) {
                return state.func_177229_b(BlockSignalBase.LAMP_STATUS);
            }
        }
        return EnumLampStatus.YELLOW_BLINKING;
    }

    private NetworkSignal<MCPos> getSignal(){
        INetworkObject<MCPos> obj = RailNetworkManager.getInstance(field_145850_b.field_72995_K).getNetwork().railObjects.get(getMCPos());
        return obj instanceof NetworkSignal ? (NetworkSignal<MCPos>)obj : null;
    }

    protected List<EntityMinecart> getNeighborMinecarts(){
        NetworkSignal<MCPos> signal = getSignal();
        if(signal == null) return Collections.emptyList();
        return getNeighborMinecarts(RailNetworkManager.getInstance(field_145850_b.field_72995_K).getNetwork().getPositionsInFront(signal).stream());
    }

    public static List<EntityMinecart> getNeighborMinecarts(Stream<MCPos> positions){
        return positions.flatMap(TileEntitySignalBase::getCartsAt).collect(Collectors.toList());
    }

    private static Stream<EntityMinecart> getCartsAt(MCPos pos){
        World world = pos.getWorld();
        return world == null ? Stream.empty() : world.func_72872_a(EntityMinecart.class, new AxisAlignedBB(pos.getPos())).stream();
    }

    @Override
    public void func_73660_a(){
        if(!field_145850_b.field_72995_K) {
            if(firstTick) {
                firstTick = false;

                setLampStatus(RailNetworkManager.getInstance(field_145850_b.field_72995_K).getLampStatus(field_145850_b, field_174879_c));
                setForceMode(RailNetworkManager.getInstance(field_145850_b.field_72995_K).getState().getForceMode(getMCPos()));
            }
        }
    }

    @Override
    public void onLoad(){
        super.onLoad();

    }

    protected void setMessage(String message, Object... arguments){
        text = message;
        this.arguments = "";
        for(int i = 0; i < arguments.length; i++) {
            if(i > 0) this.arguments += "\n";
            this.arguments += arguments[i].toString();
        }
        field_145850_b.func_184138_a(func_174877_v(), getBlockState(), getBlockState(), 3);
    }

    public String getMessage(){
        String[] localizedArguments = arguments.split("\n");
        for(int i = 0; i < localizedArguments.length; i++) {
            localizedArguments[i] = I18n.func_135052_a(localizedArguments[i]);
        }
        return I18n.func_135052_a(text, (Object[])localizedArguments);
    }

    @Override
    public void setForceMode(EnumForceMode forceMode){
        this.forceMode = forceMode;
        func_70296_d();
        if(forceMode == EnumForceMode.FORCED_GREEN_ONCE) {
            setMessage("signals.signal_message.forced_green");
        } else if(forceMode == EnumForceMode.FORCED_RED) {
            setMessage("signals.signal_message.forced_red");
        } else {
            setMessage("");
        }
    }

    @Override
    public EnumForceMode getForceMode(){
        return forceMode;
    }

    @Override
    public SPacketUpdateTileEntity func_189518_D_(){
        NBTTagCompound tag = new NBTTagCompound();
        tag.func_74778_a("text", text);
        tag.func_74778_a("arguments", arguments);
        return new SPacketUpdateTileEntity(func_174877_v(), 0, tag);
    }

    @Override
    public void onDataPacket(NetworkManager net, SPacketUpdateTileEntity pkt){
        if(pkt.func_148853_f() == 0) {
            text = pkt.func_148857_g().func_74779_i("text");
            arguments = pkt.func_148857_g().func_74779_i("arguments");
        }
    }
}
