package com.minemaarten.signals.tileentity;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.play.server.SPacketUpdateTileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ITickable;

import org.apache.commons.lang3.Validate;

import com.minemaarten.signals.api.access.IStationMarker;
import com.minemaarten.signals.capabilities.CapabilityMinecartDestination;
import com.minemaarten.signals.network.GuiSynced;
import com.minemaarten.signals.rail.network.INetworkObject;
import com.minemaarten.signals.rail.network.NetworkRail;
import com.minemaarten.signals.rail.network.PosAABB;
import com.minemaarten.signals.rail.network.mc.MCPos;
import com.minemaarten.signals.rail.network.mc.MCTrain;
import com.minemaarten.signals.rail.network.mc.RailNetworkManager;

public class TileEntityStationMarker extends TileEntityBase implements ITickable, IGUITextFieldSensitive,
        IStationMarker{
    private static int nextId;
    @GuiSynced
    private String stationName = "";
    private PosAABB<MCPos> neighborAABB;

    public TileEntityStationMarker(){
        stationName = "Station" + nextId++;
    }

    @Override
    public NBTTagCompound func_189515_b(NBTTagCompound compound){
        super.func_189515_b(compound);
        compound.func_74778_a("stationName", stationName);
        return compound;
    }

    @Override
    public void func_145839_a(NBTTagCompound compound){
        super.func_145839_a(compound);
        stationName = compound.func_74779_i("stationName");
    }

    @Override
    public String getStationName(){
        return stationName;
    }

    @Override
    public void setStationName(String stationName){
        Validate.notNull(stationName);

        this.stationName = stationName;
        if(!field_145850_b.field_72995_K) {
            RailNetworkManager.getInstance(field_145850_b.field_72995_K).markDirty(getMCPos());
            func_70296_d();
            sendUpdatePacket();
        }
    }

    public List<MCPos> getNeighborRails(){
        List<MCPos> neighbors = new ArrayList<>(1);
        for(EnumFacing d : EnumFacing.field_82609_l) {
            MCPos neighborPos = getMCPos().offset(d);
            INetworkObject<MCPos> rail = RailNetworkManager.getInstance(field_145850_b.field_72995_K).getNetwork().railObjects.get(neighborPos);
            if(rail instanceof NetworkRail) neighbors.add(rail.getPos());
        }
        return neighbors;
    }

    @Override
    public void func_73660_a(){
        if(!field_145850_b.field_72995_K) {
            updateNeighborMinecarts();
        }
    }

    public void updateNeighborMinecarts(){
        if(neighborAABB == null) {
            neighborAABB = new PosAABB<>(getNeighborRails());
        }

        RailNetworkManager.getInstance(field_145850_b.field_72995_K).getAllTrains().forEach(this::updateNeighborMinecarts);
    }

    private void updateNeighborMinecarts(MCTrain train){
        if(train.isInAABB(neighborAABB)) {
            for(EntityMinecart cart : train.getCarts()) {
                CapabilityMinecartDestination cap = cart.getCapability(CapabilityMinecartDestination.INSTANCE, null);
                if(cap.getDestinationIndex() >= 0) {
                    Pattern destinationRegex = cap.getCurrentDestinationRegex();
                    Set<MCPos> stations = RailNetworkManager.getInstance(field_145850_b.field_72995_K).getNetwork().getStations(train, destinationRegex);
                    if(stations.contains(getMCPos())) {
                        cap.nextDestination();
                    }
                }
            }
        }
    }

    @Override
    public void setText(int textFieldID, String text){
        setStationName(text);
    }

    @Override
    public String getText(int textFieldID){
        return getStationName();
    }

    @Override
    public NBTTagCompound func_189517_E_(){
        NBTTagCompound tag = super.func_189517_E_();
        tag.func_74778_a("stationName", stationName);
        return tag;
    }

    @Override
    public SPacketUpdateTileEntity func_189518_D_(){
        return new SPacketUpdateTileEntity(func_174877_v(), 0, func_189517_E_());
    }

    @Override
    public void onDataPacket(NetworkManager net, SPacketUpdateTileEntity pkt){
        super.onDataPacket(net, pkt);
        stationName = pkt.func_148857_g().func_74779_i("stationName");
    }
}
