package com.minemaarten.signals.tileentity;

import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.text.TextComponentTranslation;
import net.minecraft.world.World;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

import com.minemaarten.signals.rail.network.mc.MCPos;

public class TileEntityTeleportRail extends TileEntityRailLinkBase{
    /**
     * Only allow destination positions when:
     * 1. The position is similar to a portal position (it can't be used to cheese long distance travel)
     * 2. The receiving end has a teleport rail, to specify the destination direction.
     */
    @Override
    protected boolean isDestinationValid(MCPos destination, EntityPlayer player){
        if(func_145831_w().field_73011_w.getDimension() == destination.getDimID()) {
            player.func_145747_a(new TextComponentTranslation("signals.message.teleport_rail_failed_same_dimension", destination.getX(), destination.getY(), destination.getZ()));
            return false;
        }

        Pair<MCPos, MCPos> allowedDestinationRange = getAllowedDestinationRange(destination.getWorld());
        if(allowedDestinationRange == null) {
            player.func_145747_a(new TextComponentTranslation("signals.message.teleport_rail_failed_unloaded_destination_dimension", destination.getX(), destination.getY(), destination.getZ()));
            return false;
        } else if(destination.isInAABB(allowedDestinationRange.getLeft(), allowedDestinationRange.getRight())) {
            return true;
        } else {
            player.func_145747_a(new TextComponentTranslation("signals.message.teleport_rail_failed_invalid_location", destination.getX(), destination.getY(), destination.getZ(), allowedDestinationRange.getLeft().getX(), allowedDestinationRange.getRight().getX(), allowedDestinationRange.getLeft().getZ(), allowedDestinationRange.getRight().getZ()));
            return false;
        }
    }

    public Pair<MCPos, MCPos> getAllowedDestinationRange(World destinationDimension){
        if(destinationDimension == null) return null;

        double moveFactor = func_145831_w().field_73011_w.getMovementFactor() / destinationDimension.field_73011_w.getMovementFactor();
        double destX = MathHelper.func_151237_a(func_174877_v().func_177958_n() * moveFactor, destinationDimension.func_175723_af().func_177726_b() + 16.0D, destinationDimension.func_175723_af().func_177728_d() - 16.0D);
        double destZ = MathHelper.func_151237_a(func_174877_v().func_177952_p() * moveFactor, destinationDimension.func_175723_af().func_177736_c() + 16.0D, destinationDimension.func_175723_af().func_177733_e() - 16.0D);
        destX = MathHelper.func_76125_a((int)destX, -29999872, 29999872);
        destZ = MathHelper.func_76125_a((int)destZ, -29999872, 29999872);

        int maxDiff = 8;
        MCPos min = new MCPos(destinationDimension, new BlockPos(destX - maxDiff, 0, destZ - maxDiff));
        MCPos max = new MCPos(destinationDimension, new BlockPos(destX + maxDiff, destinationDimension.func_72940_L(), destZ + maxDiff));
        return new ImmutablePair<MCPos, MCPos>(min, max);
    }
}
